/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.application;

import cz.habi.vsp.sem.distribution.Distribution;
import cz.habi.vsp.sem.distribution.ExponentialDistribution;
import cz.habi.vsp.sem.distribution.NormalDistribution;
import cz.habi.vsp.sem.network.CrossingTarget;
import cz.habi.vsp.sem.network.FlowCrossing;
import cz.habi.vsp.sem.network.NetworkExit;
import cz.habi.vsp.sem.network.RequestGenerator;
import cz.habi.vsp.sem.network.Server;
import cz.habi.vsp.sem.network.SystemSimulation;
import cz.zcu.fav.kiv.jsim.JSimException;

/**
 * @author habi
 * 
 */
public class AppSystemSimulation {

	private Server server1;

	private Server server2;

	private Server server3;

	private Server server4;

	private RequestGenerator generatorLambda;

	private RequestGenerator generatorMu;

	private NetworkExit exit;

	private Distribution lambdaDitribution;

	private Distribution muDistribution;

	private Distribution server1Distribution;

	private Distribution server2Distribution;

	private Distribution server3Distribution;

	private Distribution server4Distribution;

	private SystemSimulation simulation;

	private int numberOfRequests;

	private double p;

	private double q;

	public AppSystemSimulation(Distribution lambdaDitribution, Distribution muDistribution, double p, double q,
			Distribution server1Distribution, Distribution server2Distribution, Distribution server3Distribution,
			Distribution server4Distribution, int numberOfRequests) throws IllegalArgumentException, JSimException {

		this.numberOfRequests = numberOfRequests;

		this.lambdaDitribution = lambdaDitribution;
		this.muDistribution = muDistribution;

		this.server1Distribution = server1Distribution;
		this.server2Distribution = server2Distribution;
		this.server3Distribution = server3Distribution;
		this.server4Distribution = server4Distribution;

		this.p = p;
		this.q = q;

		initialize();
	}

	private void initialize() throws IllegalArgumentException, JSimException {
		// simulation
		simulation = new SystemSimulation("SystemSimulation simulation", numberOfRequests);

		// generators
		generatorLambda = new RequestGenerator("LG", simulation, lambdaDitribution);
		generatorMu = new RequestGenerator("MG", simulation, muDistribution);

		// servers
		server1 = new Server("S1", simulation, server1Distribution);
		server2 = new Server("S2", simulation, server2Distribution);
		server3 = new Server("S3", simulation, server3Distribution);
		server4 = new Server("S4", simulation, server4Distribution);

		// exits
		exit = new NetworkExit(simulation);

		// connecting together
		// 1.) crossings
		FlowCrossing crossingP = new FlowCrossing(2);
		CrossingTarget ct1p = new CrossingTarget(p, server2);
		CrossingTarget ct1 = new CrossingTarget((1.0 - p), server3);
		crossingP.addDispatcherTarget(ct1p);
		crossingP.addDispatcherTarget(ct1);

		FlowCrossing crossingQ = new FlowCrossing(2);
		CrossingTarget ct2q = new CrossingTarget(q, server3);
		CrossingTarget ct2 = new CrossingTarget((1.0 - q), server4);
		crossingQ.addDispatcherTarget(ct2q);
		crossingQ.addDispatcherTarget(ct2);

		// 2.) the rest
		generatorLambda.setNextRequestAcceptor(server1);
		server1.setNextRequestAcceptor(server3);

		generatorMu.setNextRequestAcceptor(server2);
		server2.setNextRequestAcceptor(crossingP);

		server3.setNextRequestAcceptor(crossingQ);

		server4.setNextRequestAcceptor(exit);

		// activating processes
		server1.activateNow();
		server2.activateNow();
		server3.activateNow();
		server4.activateNow();

		generatorLambda.activateNow();
		generatorMu.activateNow();

		// register statistics providers
		simulation.registerStatisticsProvider(server1);
		simulation.registerStatisticsProvider(server1.getQueueStatistics());
		simulation.registerStatisticsProvider(server2);
		simulation.registerStatisticsProvider(server2.getQueueStatistics());
		simulation.registerStatisticsProvider(server3);
		simulation.registerStatisticsProvider(server3.getQueueStatistics());
		simulation.registerStatisticsProvider(server4);
		simulation.registerStatisticsProvider(server4.getQueueStatistics());
	}

	public void runSimulation(boolean latexOutput) throws JSimException {
		// simulation.runGUI();
		simulation.runSimulation();

		if (latexOutput) {
			Tools.printStatisticsLaTeX(System.out, simulation.getStatisticProviders());
		}
		else {
			Tools.printStatistics(System.out, simulation.getStatisticProviders());
		}

	}

	private static final double LAMBDA = 0.9;
	private static final double MU = 0.4;
	private static final double PROB_P = 0.5;
	private static final double PROB_Q = 0.5;
	private static final double LS1 = 1.0 / 1.0;
	private static final double LS2 = 1.0 / 0.3;
	private static final double LS3 = 1.0 / 1.0;
	private static final double LS4 = 1.0 / 0.7;
	private static final int REQUESTS = 100000;
	
	
	public static void main(String[] args) {
		System.out.println("System simulation - (c) 2007 Ivan Habernal");
		Tools.supressJSimLogging();
		
		boolean latex = false;
		
		Distribution lambdaDistribution = null;
		Distribution muDistribution = null;
		Distribution server1distribution = null;
		Distribution server2distribution = null;
		Distribution server3distribution = null;
		Distribution server4distribution = null;
		
		if (args.length == 0 || (args.length == 1 && "latex".equals(args[0]))) {
			lambdaDistribution = new ExponentialDistribution(LAMBDA);
			muDistribution = new ExponentialDistribution(MU);

			server1distribution = new ExponentialDistribution(LS1);
			server2distribution = new ExponentialDistribution(LS2);
			server3distribution = new ExponentialDistribution(LS3);
			server4distribution = new ExponentialDistribution(LS4);
			
			if (args.length == 1 && "latex".equals(args[0])) {
				latex = true;
			}
		}
		else {
			double c = 0.1;
			if ("N1".equals(args[0])) {
				// C = 0.05
				c = 0.05;
			}
			else if ("N2".equals(args[0])) {
				// C = 0.3
				c = 0.3;
			}
			else if ("N3".equals(args[0])) {
				// C = 0.8
				c = 0.8;
			}
			else {
				System.out.println("Unknown command line parameters.");
				System.out.println("$(java run-this) [ (no-param) | N1 | N2 | N3 ]");
				System.exit(1);
			}
			
			lambdaDistribution = new NormalDistribution(LAMBDA, LAMBDA * c);
			muDistribution = new NormalDistribution(LAMBDA, LAMBDA * c);

			server1distribution = new NormalDistribution(LAMBDA, LAMBDA * c);
			server2distribution = new NormalDistribution(LAMBDA, LAMBDA * c);
			server3distribution = new NormalDistribution(LAMBDA, LAMBDA * c);
			server4distribution = new NormalDistribution(LAMBDA, LAMBDA * c);
			
			if ("latex".equals(args[args.length - 1])) {
				latex = true;
			}
		}
		
		if (latex) {
			System.out.println("LaTeX mode enabled");
		}
		
		try {
			AppSystemSimulation appSystemSimulation = new AppSystemSimulation(lambdaDistribution, muDistribution, PROB_P, PROB_Q,
					server1distribution, server2distribution, server3distribution, server4distribution, REQUESTS);
			
			System.out.println("System created, simulation is running...");
			appSystemSimulation.runSimulation(latex);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (JSimException e) {
			e.printStackTrace();
		}

	}
}
