/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.network;

import java.util.LinkedHashSet;
import java.util.Set;

import cz.habi.vsp.sem.distribution.Distribution;
import cz.habi.vsp.sem.distribution.UniformDistribution;
import cz.zcu.fav.kiv.jsim.JSimException;

/**
 * This class represents the network crossing - a single (or multiple) input and
 * multiple outputs network element. It contais set of targets, each target with
 * its probability.
 * 
 * @author habi
 * 
 */
public class FlowCrossing implements RequestAcceptor {

	private Set<CrossingTarget> targets;

	private int capacity;

	private Distribution distribution = new UniformDistribution(0, 1);

	/**
	 * Creates a new instance of FlowCrossing
	 * 
	 * @param capacity
	 *            capacity of this crossing (count of RequestAcceptors which
	 *            this crossing leads to)
	 * @throws IllegalArgumentException
	 *             if the capacity is negative
	 */
	public FlowCrossing(int capacity) throws IllegalArgumentException {
		if (capacity <= 0) {
			throw new IllegalArgumentException("Capacity must greater than zero");
		}

		this.capacity = capacity;

		targets = new LinkedHashSet<CrossingTarget>(capacity);
	}

	/**
	 * Adds the dispatcher target.
	 * 
	 * @param crossingTarget
	 *            target
	 * @throws IllegalStateException
	 *             if this crossing is already full (all targets has been set,
	 *             according to the capacity)
	 * @throws IllegalArgumentException
	 *             if the crossingTarget is null
	 */
	public void addDispatcherTarget(CrossingTarget crossingTarget) throws IllegalStateException, IllegalArgumentException {
		if (targets.size() >= capacity) {
			throw new IllegalStateException("All targets has already been set");
		}

		if (crossingTarget == null) {
			throw new IllegalArgumentException("crossing target is null");
		}

		targets.add(crossingTarget);

		// we have all targets -> validate probabilities
		if (targets.size() == capacity) {
			if (!validateTargets()) {
				throw new IllegalArgumentException("The sum of probabilities of all target is not 1");
			}
		}
	}

	private static final double EPSILON = 0.0001;

	/**
	 * Returns true, if the sum of probabilites of all targets equals to 1.
	 * 
	 * @return true, if sum == 1
	 */
	private boolean validateTargets() {
		double result = 0.0;

		for (CrossingTarget ct : targets) {
			result += ct.getPropability();
		}

		return Math.abs(result - 1.0) < EPSILON;
	}

	/**
	 * Returns the next requestAcceptor. This method uses the probability of
	 * crossingTarget and the random number from crossing's distribution.
	 * 
	 * @return next request acceptor
	 */
	private RequestAcceptor getNextRequestAcceptor() {
		double random = distribution.getRandomValue();

		double sum = 0.0;

		for (CrossingTarget ct : targets) {
			sum += ct.getPropability();

			if (random < sum) {
				return ct.getRequestAcceptor();
			}
		}

		throw new IllegalStateException("We should have found a requestAcceptor!");
	}

	public void acceptRequest(Request request) throws JSimException {
		if (capacity != targets.size()) {
			throw new IllegalStateException("Capacity of this crossing is " + capacity + ", but only " + targets.size()
					+ " targets have been set");
		}

		// deliver the request to the next requestAcceptor
		getNextRequestAcceptor().acceptRequest(request);
	}

}
