/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.network;

import cz.habi.vsp.sem.distribution.Distribution;
import cz.zcu.fav.kiv.jsim.JSimException;
import cz.zcu.fav.kiv.jsim.JSimInvalidParametersException;
import cz.zcu.fav.kiv.jsim.JSimProcess;
import cz.zcu.fav.kiv.jsim.JSimSecurityException;
import cz.zcu.fav.kiv.jsim.JSimSimulation;

/**
 * This abstract class represents a network process which can do some work with
 * request and deliver it to the next queue element.
 * 
 * @author habi
 * 
 */
public abstract class NetworkProcess extends JSimProcess implements Connectable {

	private Distribution distribution;

	/**
	 * Creates a new instance of NetworkProcess
	 * 
	 * @param name
	 *            name of the process
	 * @param simulation
	 *            the simulation
	 * @param distribution
	 *            the distribution of probability
	 * @throws JSimException
	 *             if an error occurs while creating JSimProcess
	 * @throws IllegalArgumentException
	 *             if the distribution is null
	 */
	public NetworkProcess(String name, JSimSimulation simulation, Distribution distribution) throws JSimException, IllegalArgumentException {
		super(name, simulation);
		if (distribution == null) {
			throw new IllegalArgumentException("Distribution is null");
		}
		
		this.distribution = distribution;
	}

	/**
	 * Resturn the distribtution of this process.
	 * 
	 * @return distribution
	 */
	public Distribution getDistribution() {
		return distribution;
	}

	/**
	 * This is the main body of {@link NetworkProcess}. This method is called
	 * every time the {@link NetworkProcess} wakes up to do some work.
	 * 
	 * @throws JSimException
	 *             if an exception occurs
	 */
	protected abstract void doWork() throws JSimException;

	/**
	 * Returns false, if the life cycle of this {@link NetworkProcess} is
	 * finished (true, if it should do some work, respectively).
	 * 
	 * @return boolean value
	 */
	protected abstract boolean workingCondition();

	/**
	 * The {@link NetworkProcess} is held for some random time (depends on the
	 * random value retrieved from {@link NetworkProcess}'s distribution). If
	 * the random number is less than zero, the process sleeps for 'zero-time'.
	 * 
	 * @return time the process has been sleeping for
	 * @throws JSimException
	 *             if an error occurs
	 */
	protected double holdForRandomTime() throws JSimException {
		double result = distribution.getRandomValue();

		if (result < 0) {
			result = 0;
		}

		hold(result);
		return result;
	}

	/**
	 * This method contains the infinite-loop. In every cycle of this loop it's
	 * beeing tested, whether the {@link NetworkProcess#workingCondition()} is
	 * true and the {@link NetworkProcess#doWork()} is called.
	 */
	@Override
	protected void life() {
		try {
			while (workingCondition()) {
				// we do some work... :)
				doWork();
			}
		} catch (JSimSecurityException e) {
			e.printStackTrace();
		} catch (JSimInvalidParametersException e) {
			e.printStackTrace();
		} catch (JSimException e) {
			e.printStackTrace();
		}
	}

	private RequestAcceptor nextRequestAcceptor;

	public RequestAcceptor getNextRequestAcceptor() {
		return nextRequestAcceptor;
	}

	public void setNextRequestAcceptor(RequestAcceptor requestAcceptor) throws IllegalArgumentException {
		if (requestAcceptor == null) {
			throw new IllegalArgumentException("requestAcceptor is null");
		}
		this.nextRequestAcceptor = requestAcceptor;
	}
}
