/**
 * Hlavní semestrální práce z KIV/VSP
 * (c) 2006 - 2007 Ivan Habernal
 * e-mail: habernal@students.zcu.cz
 */
package cz.habi.vsp.sem.network;

import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.Map;

import org.apache.commons.math.stat.descriptive.SummaryStatistics;

/**
 * This class represents a queue for requests. It also provides some statistics.
 * 
 * @author habi
 * 
 */
public class Queue implements StatisticsProvider {

	private SystemSimulation parentSimulation;

	private SummaryStatistics lwStatistics = SummaryStatistics.newInstance();

	private SummaryStatistics twStatistics = SummaryStatistics.newInstance();

	private Server parentServer;
	
	// starting time
	private double startingTime;
	
	// temp value for computing Lw
	private double lwTemp = 0;
	
	private double lastTimeChange;

	/**
	 * Creates a new instance of Queue
	 * 
	 * @param parentSimulation
	 *            parent instance JSimSimulation
	 * @param parentServer
	 *            parent server
	 * @throws IllegalArgumentException
	 *             if the parent simulation or parent server is null
	 */
	public Queue(SystemSimulation parentSimulation, Server parentServer) throws IllegalArgumentException {
		if (parentSimulation == null) {
			throw new IllegalArgumentException("Parent simulation is null");
		}
		if (parentServer == null) {
			throw new IllegalArgumentException("Parent server is null");
		}
		
		this.parentSimulation = parentSimulation;
		this.parentServer = parentServer;
		
		this.startingTime = parentSimulation.getCurrentTime();
		this.lastTimeChange = startingTime;
	}

	private LinkedList<Request> waitingRequests = new LinkedList<Request>();

	/**
	 * Adds the request to the queue.
	 * 
	 * @param request
	 *            request
	 * @throws IllegalArgumentException
	 *             if the request is null
	 */
	public void addRequestToQueue(Request request) throws IllegalArgumentException {
		updateLwStats();

		waitingRequests.add(request);
		request.enterQueue(parentSimulation.getCurrentTime());

		// updating statistics
		lwStatistics.addValue(waitingRequests.size());
		
	}
	
	/**
	 * Updates statitics for Lw
	 */
	private void updateLwStats() {
		int currentSize = waitingRequests.size();
		double interval = parentSimulation.getCurrentTime() - lastTimeChange;
		lwTemp += (interval * currentSize);
		
		this.lastTimeChange = parentSimulation.getCurrentTime();
	}

	/**
	 * Returns the next request from the queue.
	 * 
	 * @return next request or null, if the queue is empty
	 */
	public Request getNextRequest() {
		if (waitingRequests.isEmpty()) {
			return null;
		}
		updateLwStats();

		Request result = waitingRequests.removeFirst();
		result.leaveQueue(parentSimulation.getCurrentTime());

		// update statistics
		lwStatistics.addValue(waitingRequests.size());
		
		twStatistics.addValue(result.getWaitingTimeInLastQueue());

		return result;
	}
	
	/**
	 * Returns L<sub>W</sub> - the average queue length.
	 * 
	 * @return average length
	 */
	public double getAverageQueueLength() {
		double totalWorkingTime = parentSimulation.getCurrentTime() - startingTime;
		double avgQueueLength = lwTemp / totalWorkingTime;
		
		return avgQueueLength;
	}

	/**
	 * Returns T<sub>W</sub> - the average request's waiging time in the
	 * queue.
	 * 
	 * @return average waiting time
	 */
	public double getAverageWaitingTime() {
		return twStatistics.getMean();
	}
	
	public double getDeviationWaitingTime() {
		return twStatistics.getStandardDeviation();
	}

	public String getProviderName() {
		return "(Queue) before " + parentServer.getProviderName();
	}

	public Map<String, Object> getStatistics() {
		Map<String, Object> map = new LinkedHashMap<String, Object>();
		
		map.put("Tw", getAverageWaitingTime());
		map.put("Tw (DEV)", getDeviationWaitingTime());
		map.put("Lw", getAverageQueueLength());
		
		return map;
	}
}
